"""
This module implements functionc needed to do curvefitting and display the output.
TODO: check statsmodel library of python for the same
(http://statsmodels.sourceforge.net/documentation.html)
"""
import math
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
import constants
from table_processing import TableProcessing


def residue(obs, pred):
    """
    Calculate residual from observed and fitted values
    Args:
        obs: list of observed values
        pred: list of predicted values
    Returns:
        error: difference of observed and predicted values
    """
    error = obs - pred
    return error


def sst(obs, y_av, weights=None):
    """
    Sum of squares about the mean

    Args:
        obs: list of observed values
        y_av: average of observed values
        weights: weights to apply to the y-coordinates of the sample points.
        For gaussian uncertainties, use sigma (not 1/sigma**2).

    Returns:
        sum_of_squares: Summation(y-y_av)^2/weight^2
    """
    error = obs - y_av
    if weights is not None:
        return ((error ** 2) / (weights ** 2)).sum()
    return ((error ** 2).sum())


def sse(obs, pred, weights=None):
    """
    sum of squares of residuals

    Args:
    obs : list of observed values
    pred : list of predicted values
    weights: weights to apply to the y-coordinates of the sample points.
    For gaussian uncertainties, use sigma (not 1/sigma**2).

    Returns:
    sum_of_squares: Summation(y_obs-y_pred)^2/weight^2
    """

    error = obs - pred
    if weights is not None:
        sum_of_squares = ((error ** 2) / weights ** 2).sum()
    else:
        sum_of_squares = (error ** 2).sum()

    return sum_of_squares


def r_square(sse, sst):
    """
    R squared value is "fraction of variance explained by the model".
    We are using the most generic formula 1 - sse/sst to calculate R^2 which
    is used by Excel. The other package used to verify our curve fit module is
    lm (https://stat.ethz.ch/R-manual/R-devel/library/stats/html/lm.html)in R.
    Now the formula for for R^2 in this package is
    (https://stat.ethz.ch/R-manual/R-devel/library/stats/html/summary.lm.html):
    R^2 = 1 - Sum(R[i]^2) / Sum((y[i]- y*)^2),
    where y* is the mean of y[i] if there is an intercept and zero otherwise.
    So the R^2 values are different in R and Excel if there is no intercept.
    These are some links which explain why model with no intercept should be
    avoided unless sure about it and also these explain how these cause apparent
    increase in R^2 values which might not be statistically significant
    http://statistiksoftware.blogspot.in/2013/01/why-we-need-intercept.html?m=1,
    http://stats.stackexchange.com/questions/26176/removal-of-statistically-
    significant-intercept-term-increases-r2-in-linear-mo
    http://www.ats.ucla.edu/stat/mult_pkg/faq/general/noconstant.htm
    Args:
        sse: sum of squares of residuals
        sst: total sum of squares
    Returns:
        r_squared value
    """
    if sse == 0 and sst == 0:
        return 0
    return 1 - sse / sst

def rmse(sse, deg_freedom):
    """
    Root mean square error

    Args:
        sse: sum of squares of residuals
        deg_freedom: The model degrees of freedom, is the length
        of y minus the number of parameters calculated in the model.

    Returns:
        Squareroot(Summation(y_obs - y_pred)^2/n)
        """
    if sse == 0:
        return 0
    return (sse / deg_freedom) ** 0.5

def loglog_no_intercept(x, k):
    """
    power model in form ax^k
    Args:
        x: variable
        a: constant of multiplication term
        k: constant for power term
    Returns:
        ax^k
    """
    for n, i in enumerate(x):
        if i <= 0:
            x[n] = 1.0
    return k*np.log(x)

def loglog_intercept(x, a, k):
    """
    power model in form ax^k
    Args:
        x: variable
        a: constant of multiplication term
        k: constant for power term
    Returns:
        ax^k
    """
    for n, i in enumerate(x):
        if i <= 0:
            x[n] = 1.0
    return a + k*np.log(x)

def linear_no_intercept(x, m):
    """
    linear model with slope m and constant c
    """
    return np.multiply(m, x)


def linear_intercept(x, m, c):
    """
    linear model with slope m and constant c
    """
    return np.multiply(m, x) + c


def exp1(x, a, b, c):
    """
    exponential model with one exp term and constant
    MATLAB exp1 has no intercept but we need it to cover the
    case with x=0,y=0

    Args:
        x: variable raised to exp
        a: coefficient of exp term
        b: coefficient of x
        c: constant term

    Returns:
        ae^(bx) + c
    """
    return a * np.exp(np.multiply(-b, x)) + c


def exp2(x, a, b, c, d):
    """
    exponential model with two exp terms and constant
    MATLAB exp2 has the same model, so not adding intercept.
    This model allows x=0,y=0 condition possible, which might
    be a possibility in the scenario where this package is used, no
    intercept needed to cover that case.

    Args:
        x: variable raised to exp
        a: coefficient of first exp term
        b: coefficient of first x
        c: coefficient of second exp term
        d: coefficient of second x
        e: constant term

    Returns:
        ae^(-bx) + ce^(-dx)
    """
    return a * np.exp(np.multiply(-b, x)) + c * np.exp(np.multiply(-d, x))


def loglogtranformation(x, base):
    """
    Take log of given float value wrt base. If 0 return 0
    """
    if x != 0:
        return math.log(x, base)
    elif x == 0:
        return 0

def inverselogtranform(x, base):
    """
    Assuming that if x = 0 that's due to to initial value 1 and not 0
    (it can be 0 if the quantity had gone through loglogtranformation)
    Preferably use for constants and not for transformed intensities/
    concentrations
    """
    return np.power(base, x)

def power_func(x, a, k):
    """
    power model in form ax^k
    Args:
        x: variable
        a: constant of multiplication term
        k: constant for power term
    Returns:
        ax^k
    """
    return np.multiply(a, np.power(x, k))

def goodness_of_fit(ydata, yfit, num_model_params, weights):
    """
    Calculate various goodness of fit parameters

    Args:
        ydata: observed y values
        yfit: fitted y values
        weights:weights to apply to the y-coordinates of the sample points.
        For gaussian uncertainties, use sigma (not 1/sigma**2)

    Returns:
        res: residues
        sse_lin: sum of squares of residuals
        r_square_lin: r squared values, these should be considered carefully
        while judging non linear models
        rmse_lin: root mean squared error
    """
    if weights is not None:
        weights = [float(x) for x in weights]
        weights = np.array(weights)
    ydata = np.array(ydata)
    yfit = np.array(yfit)

    res = residue(ydata, yfit)
    sse_val = sse(ydata, yfit, weights)
    sst_val = sst(ydata, np.mean(ydata), weights)
    r_square_val = r_square(sse_val, sst_val)
    rmse_val = rmse(sse_val, len(ydata) - num_model_params)
    gof_dict = {'sse': sse_val, 'r_square': r_square_val, 'rmse': rmse_val}
    #Replacing infinite values fit parameters with NaN
    updated_values = dict((key, np.NaN)
                          for key, value in gof_dict.iteritems() if math.isinf(value))
    gof_dict.update(updated_values)

    return res, gof_dict.get('sse'), gof_dict.get('r_square'), gof_dict.get('rmse')


def curvefit_linear(xdata, ydata, intercept, weights=None):
    """
    Fitting the linear model, returns the fitted values
    and goodness of fit

    Args:
        xdata: x data for the curve fit
        ydata: y data for the curve fit
        intercept: boolean values corresponding to intercept
        weights: weights to apply to the y-coordinates of the sample points.
        For gaussian uncertainties, use sigma (not 1/sigma**2).

    Returns:
        yfit: fitted y values
        popt: fitted parameters
        res: residues
        sse_lin: sum of squares of residuals
        r_square_lin: r squared values
        rmse_lin: root mean squared error
    """
    # TODO: Test for curvefit intercept
    if weights is not None:
        weights = [float(x) for x in weights]
        if intercept:

            popt, pcov = curve_fit(linear_intercept, xdata, ydata,
                                   sigma=np.array(weights))
        else:
            popt, pcov = curve_fit(linear_no_intercept, xdata, ydata,
                                   sigma=np.array(weights))
    else:
        if intercept:
            popt, pcov = curve_fit(linear_intercept, xdata, ydata)
        else:
            popt, pcov = curve_fit(linear_no_intercept, xdata, ydata)
    if intercept:
        yfit = linear_intercept(np.array(xdata), popt[0], popt[1])
    else:
        yfit = linear_no_intercept(np.array(xdata), popt[0])
    res, sse_lin, r_square_lin, rmse_lin = goodness_of_fit(ydata, yfit, 1,
                                                           weights)
    return yfit, popt, res, sse_lin, r_square_lin, rmse_lin


def loglogtranform_numpyarray(xdata, base):
    """
    loglogtranform of a numpyarray where log(0) is returned to 0
    """
    return np.array([loglogtranformation(x, base) for x in xdata])


def curvefit_poly(xdata, ydata, deg=2, weights=None):
    """
    Fitting the polynomial model, returns the fitted values and goodness of fit
    TODO: Implement orthogonal polynomial as in R
    (http://stackoverflow.com/questions/29999900/poly-in-lm-difference-between-
    raw-vs-orthogonal)

    Args:
        xdata: x data for the curve fit
        ydata: y data for the curve fit
        deg: degree of the polynomial to be fit
        weights: weights to apply to the y-coordinates of the sample points.
        For gaussian uncertainties, use sigma (not 1/sigma**2).

    Returns:
        yfit: fitted y values
        popt: fitted parameters
        res: residues
        sse_lin: sum of squares of residuals
        r_square_lin: r squared values
        rmse_lin: root mean squared error
    """
    if weights is not None:
        weights = [float(x) for x in weights]
    else:
        weights = len(ydata) * [1]
    popt = np.polyfit(xdata, ydata, deg, w=1 / np.array(weights))
    yfit = np.polyval(popt, np.array(xdata))
    res, sse_poly, r_square_poly, rmse_poly = goodness_of_fit(ydata, yfit,
                                                              deg + 1, weights)
    return yfit, popt, res, sse_poly, r_square_poly, rmse_poly


def curvefit_exp(xdata, ydata, deg=1, weights=None):
    """
    Fitting the exponential model, returns the fitted values
    and goodness of fit.

    TODO: Add initial conditions as args, see lm-fit python library.

    Args:
        xdata: x data for the curve fit
        ydata: y data for the curve fit
        deg: number of exponential terms in the model
        weights: weights to apply to the y-coordinates of the sample points.
        For gaussian uncertainties, use sigma (not 1/sigma**2).

    Returns:
        yfit: fitted y values
        popt: fitted parameters
        res: residues
        sse_lin: sum of squares of residuals
        r_square_lin: r squared values
        rmse_lin: root mean squared error
    """
    if weights is not None:
        weights = [float(x) for x in weights]
    else:
        weights = len(ydata) * [1]

    if deg == 1:
        popt, pcov = curve_fit(exp1, xdata, ydata, sigma=np.array(weights))
    else:
        popt, pcov = curve_fit(exp2, xdata, ydata, sigma=np.array(weights))
    if deg == 1:
        yfit = exp1(np.array(xdata), *popt)
        res_exp, sse_exp, r_square_exp, rmse_exp = goodness_of_fit(ydata, yfit,
                                                                   3, weights)
    else:
        yfit = exp2(np.array(xdata), *popt)
        res_exp, sse_exp, r_square_exp, rmse_exp = goodness_of_fit(ydata, yfit,
                                                                   4, weights)
    return yfit, popt, res_exp, sse_exp, r_square_exp, rmse_exp


def curvefit_pow(xdata, ydata, weights=None):
    """
    curvefitting for power function

    Args:
        xdata: x data for the curve fit
        ydata: y data for the curve fit
        weights: weights to apply to the y-coordinates of the sample points

    Returns:
        yfit: fitted y values
        popt: fitted parameters
        res_pow: residues
        sse_pow: sum of squares of residuals
        r_square_pow: r squared values
        rmse_pow: root mean squared error
    """
    if weights is not None:
        weights = [float(x) for x in weights]
    else:
        weights = len(ydata) * [1]

    popt, pcov = curve_fit(power_func, xdata, ydata, sigma=np.array(weights),maxfev=2000)
    yfit = power_func(np.array(xdata), *popt)
    res_pow, sse_pow, r_square_pow, rmse_pow = goodness_of_fit(ydata, yfit, 2,
                                                               weights)
    return yfit, popt, res_pow, sse_pow, r_square_pow, rmse_pow


def curvefit_loglog(xdata, ydata, intercept, weights=None):
    """
    curvefitting for loglog function

    Args:
        xdata: x data for the curve fit
        ydata: y data for the curve fit
        weights: weights to apply to the y-coordinates of the sample points

    Returns:
        yfit: fitted y values
        popt: fitted parameters
        res_log: residues
        sse_log: sum of squares of residuals
        r_square_log: r squared values
        rmse_log: root mean squared error
    # """
    if weights is not None:
        weights = [float(x) for x in weights]
        if intercept:
            popt, pcov = curve_fit(loglog_intercept, xdata, ydata,
                                   sigma=np.array(weights))
        else:
            popt, pcov = curve_fit(loglog_no_intercept, xdata, ydata,
                                   sigma=np.array(weights))
    else:
        if intercept:
            popt, pcov = curve_fit(loglog_intercept, xdata, ydata)
        else:
            popt, pcov = curve_fit(loglog_no_intercept, xdata, ydata)
    if intercept:
        yfit = loglog_intercept(np.array(xdata), popt[0], popt[1])
    else:
        yfit = loglog_no_intercept(np.array(xdata), popt[0])
    res_log, sse_log, r_square_log, rmse_log = goodness_of_fit(ydata, yfit, 1,
                                                           weights)
    return yfit, popt, res_log, sse_log, r_square_log, rmse_log

def curvefit(xdata, ydata, intercept, type='linear', degree=1, weights=None):
    """
    Curve fitting for different types of models

    Args:
        xdata: x data for the curve fit
        ydata: y data for the curve fit
        intercept: boolean values corresponding to intercept
        type: type of model, can be linear,exp, poly, loglog or power
        degree: degree for polynomial model and exponential, always 1 for linear
        weights: weights to apply to the y-coordinates of the sample points.
        For gaussian uncertainties, use sigma (not 1/sigma**2).

    Returns:
        yfit: fitted y values
        popt: fitted parameters
        res: residues
        sse: sum of squares of residuals
        r_square: r squared values
        rmse: root mean squared error
    """
    if type == 'linear':
        return curvefit_linear(xdata, ydata, intercept, weights=weights)
    elif type == 'exp':
        return curvefit_exp(xdata, ydata, degree, weights)
    elif type == 'poly':
        return curvefit_poly(xdata, ydata, degree, weights)
    elif type == 'loglog':
        return curvefit_loglog(xdata, ydata, intercept, weights=weights)
    elif type == 'power':
        return curvefit_pow(xdata, ydata, weights=weights)


def interpolation(intensities, popt, ftype, intercept):
    """
    get y (conc) values given x (intensities) in a model

    Args:
        intensities: input x values
        popt: parameters of the model eg for linear model ax+b,
        popt will be [a,b]
        ftype: model type, linear, exp, poly, loglog or power
        intercept: boolean values corresponding to intercept

    Returns:
        conc_list: output y values
    """

    if ftype == 'linear':
        if intercept:
            conc_list = linear_intercept(intensities, popt[0], popt[1])
        else:
            conc_list = linear_no_intercept(intensities, popt[0])
    elif ftype == 'exp':
        if len(popt) == 3:
            conc_list = exp1(intensities, *popt)
        elif len(popt) == 4:
            conc_list = exp2(intensities, *popt)
    elif ftype == 'poly':
        conc_list = np.polyval(popt, intensities)
    elif ftype == 'loglog':
        if intercept:
            conc_list = loglog_intercept(intensities, popt[0], popt[1])
        else:
            conc_list = loglog_no_intercept(intensities, popt[0])
    elif ftype == 'power':
        conc_list = power_func(intensities, *popt)
    else:
        conc_list = []
    return conc_list


def intensity_to_conc(metab, biosamples, unit, popt, type, intercept):
    """
    Convert intensity to concentration in biosamples dataframe for a metabolite

    Args:
        metab: name of the unlabeled fragment for which conc is to be calculated
        biosamples: dataframe containing experiment samples, for who intensity values
        are to be converted to concentration
        unit: unit of biosamples concentration
        popt: parameters of the model ex for linear model ax+b,
        popt will be [a,b]
        type: model type, linear, exp or poly
        intercept: boolean values corresponding to intercept

    Returns:
        biosamples: biosamples dataframe with new concentration values
    """
    biosamples.insert(len(biosamples.columns), constants.UNIT, unit,
                      allow_duplicates=True)
    biosamples.ix[
        biosamples[constants.PARENT] == metab, constants.MQ_CONCENTRATION] = \
        interpolation(biosamples.ix[biosamples[
                                        constants.PARENT] == metab,
                                    constants.MQ_INTENSITY],popt, type, intercept)
    return biosamples


def rounding_for_display(con):
    """
    case when the number if less than .001 representing them in terms of
    scientific notation
    """
    if np.around(con, 3) == 0:
        strtoapp = '%.2E' % (con)
    else:
        if math.floor(con) == con:
            strtoapp = '%.1f' % (con)
        else:
            strtoapp = '%.4f' % (con)

    return strtoapp


def equation_sign(strtoapp, rounded_num, popt, pos):
    """
    displaying proper sign of the equation
        Args:
        strtoapp: absolute value of number to be displayed
        rounded_num: actual number to get the sign
        popt: list of parameters to be displayed
        pos: position of the number in the equation
        Returns:
        strtoapp: number with appropriate sign
    """
    if rounded_num > 0:
        # this is to avoid the + sign in the start if the equation
        if len(popt) - 1 == pos:
            strtoapp = strtoapp
        else:
            # this is for other cases other than the first element of the
            # equation
            strtoapp = '+ ' + strtoapp
    else:
        # this is to avoid the - sign in the start if the equation
        if len(popt) - 1 == pos:
            strtoapp = '-' + strtoapp
        else:
            # this is for other cases other than the first element of the
            # equation
            strtoapp = '- ' + strtoapp

    return strtoapp


def equation_to_string(popt, type, intercept):
    """
    convert equation to string for display
    Args:
        popt: parameters of the equation to be displayed
        type: type of model, can be linear, poly, exp, loglog or power
        intercept: boolean values corresponding to intercept
    Returns:
        equation_string: string form of the equation according to the model
    """
    equation_string = ''
    if type == 'linear':
        if intercept:
            equation_string = equation_sign(
                rounding_for_display(math.fabs(popt[0])), popt[0], popt,
                1) + '*x' + \
                              equation_sign(
                                  rounding_for_display(math.fabs(popt[1])),
                                  popt[1], popt, 2)
        else:
            equation_string = equation_sign(
                rounding_for_display(math.fabs(popt[0])), popt[0], popt,
                0) + '*x'

    if type == 'poly':
        part = ''
        n = len(popt)
        for con in popt:
            n = n - 1
            if n != 0:
                strtoapp = rounding_for_display(math.fabs(con)) + '*x^%i' % (n)
            else:
                strtoapp = rounding_for_display(math.fabs(con))

            strtoapp = equation_sign(strtoapp, con, popt, n)

            part = part + strtoapp
        equation_string = part
    if type == 'exp':
        n = len(popt)
        if n == 3:
            equation_string = equation_sign(
                rounding_for_display(math.fabs(popt[0])), popt[0], popt, 2) \
                              + '*e^{(' + str(
                rounding_for_display(popt[1])) + 'x)} ' \
                              + equation_sign(
                rounding_for_display(math.fabs(popt[2])), popt[2], popt, 0)
        if n == 4:
            equation_string = equation_sign(
                rounding_for_display(math.fabs(popt[0])), popt[0], popt, 3) \
                              + '*e^{(' + str(
                rounding_for_display(popt[1])) + 'x)} ' \
                              + equation_sign(
                rounding_for_display(math.fabs(popt[2])), popt[2], popt, 1) \
                              + '*e^{(' + str(
                rounding_for_display(popt[3])) + 'x)} '
    if type == 'loglog':
        if intercept:
            equation_string = equation_sign(
                rounding_for_display(math.fabs(popt[0])), popt[0], popt,
                1) + '*log(x) ' + \
                              equation_sign(
                                  rounding_for_display(math.fabs(popt[1])),
                                  popt[1], popt, 0)
        else:
            equation_string = equation_sign(
                rounding_for_display(math.fabs(popt[0])), popt[0], popt,
                0) + '*log(x) '
    if type == 'power':
        equation_string = equation_sign(
            rounding_for_display(math.fabs(popt[0])), popt[0], popt,
            1) + '*x^{(' + \
                          equation_sign(
                              rounding_for_display(math.fabs(popt[1])), popt[1],
                              popt, 1) + ')} '

    return equation_string


def get_fit_type(fitType):
    """
    finds out the fit_type needed for the curvefit

    It reads fitType parameter and finds out the fitting
    function needed for particular curvefit

    Args:
        fitType: object containing PLOTTYPE and degree

    Returns:
        fType: method needed for curvefit
    """
    if fitType[constants.PLOTTYPE] == constants.LINEAR:
        fType = constants.LIN_PKG
        fitType[constants.DEGREE] = 1
    elif fitType[constants.PLOTTYPE] == constants.POLY:
        fType = constants.POLY_PKG
    elif fitType[constants.PLOTTYPE] == constants.EXP:
        fType = constants.EXP_PKG
    elif fitType[constants.PLOTTYPE] == constants.LOGLOG:
        fType = constants.LOGLOG_PKG
        fitType[constants.DEGREE] = 1
    elif fitType[constants.PLOTTYPE] == constants.POWER:
        fType = constants.POWER_PKG
        fitType[constants.DEGREE] = 1
    return fType


def get_weight(dct, index, std_table_metab):
    """
    Decides the weight corresponding to info provided from paramater_json

    This function checks the input provided from paramater_json and finds
    out the weights for each of the metabolite and returns it.

    Args:
        dct: variable passing while iterating over list
        index: index of variable
        std_table_metab: standard table of the metabolite

    Returns:
        weight: weight of all standard points
    """

    if dct[constants.WEIGHT] == constants.NO_WEIGHT:
        weight = 1.
    else:
        if dct[constants.WEIGHT] == '1/variance':
            if std_table_metab[constants.STD_DEV][index] == float(0):
                weight = 1.
            else:
                weight = std_table_metab[constants.STD_DEV][index]
        else:
            if dct[constants.CONC] == float(0):
                weight = 1.
            else:
                weight = (dct[constants.CONC]) ** 0.5
    return weight


def create_curvefit_info(metab_name, standard_data, fitType, merged_std_df,
                         intercept, background_corr):
    """
    This method returns the parameters for the curvefit
    Args:
        metab_name: name of the metabolite for which we have
                    standard data
        standard_data: Standard samples data.
        fitType: object containing PlotType and degree
        merged_std: df containing std samples mapped to corresponding
                    conc
        intercept: boolean values corresponding to intercept
    Returns:
            yfit: fitted y values
            popt: fitted parameters
            res: residues
            sse: sum of squares of residuals
            r_square: r squared values
            rmse: root mean squared error
            intensities: average area of standard samples for corresponding conc
            fType: type of model needed for create curvefit, can be linear,
                    exp, poly, loglog or power
    """
    fType = get_fit_type(fitType)

    intensities = []
    concentrations = []
    weights = []
    table_processing = TableProcessing(merged_std_df)
    standard_table, y_unit = table_processing.get_all_std_tables_df(background_corr)
    std_table_metab = standard_table[
        standard_table[constants.PARENT] == metab_name].reset_index()

    standard_data_index = 0
    for standard_data_dictionary in standard_data:
        if not standard_data_dictionary.get(constants.REJECT):
            intensities.append(standard_data_dictionary.get(constants.AVGAREA))
            concentrations.append(standard_data_dictionary.get(constants.CONC))
            weight = get_weight(standard_data_dictionary, standard_data_index, std_table_metab)
            weights.append(weight)
        standard_data_index = standard_data_index + 1
    yfit, popt, res, sse, r_square, rmse = curvefit(intensities, concentrations,
                                                    intercept, fType,
                                                    int(fitType[
                                                            constants.DEGREE]),
                                                    weights)
    return yfit, popt, res, sse, r_square, rmse, intensities, fType, y_unit


def create_best_curvefit_suggested_info(metab_name, standard_data, fitType, merged_std_df,
                         intercept, background_corr):
    """
        This function calculates the best plot for each type of graph
        by removing best suitable standard.
        We calculate the distance for each standard from the curve and
        then remove the standard with maximum distance and recaluculate
        R_square value. If its better then remove else nothing.
    Args:
        metab_name: name of the metabolite for which we have
                    standard data
        standard_data: Standard samples data.
        fitType: object containing PlotType and degree
        merged_std: df containing std samples mapped to corresponding
                    conc
        intercept: boolean values corresponding to intercept
    Returns:
            yfit: fitted y values
            popt: fitted parameters
            res: residues
            sse: sum of squares of residuals
            r_square: r squared values
            rmse: root mean squared error
            intensities: average area of standard samples for corresponding conc
            fType: type of model needed for create curvefit, can be linear,
                    exp, poly, loglog or power
    """
    fType = get_fit_type(fitType)

    intensities = []
    concentrations = []
    weights = []
    table_processing = TableProcessing(merged_std_df)
    standard_table, y_unit = table_processing.get_all_std_tables_df(background_corr)
    std_table_metab = standard_table[
        standard_table[constants.PARENT] == metab_name].reset_index()

    standard_data_index = 0
    for standard_data_dictionary in standard_data:
        if not standard_data_dictionary.get(constants.REJECT):
            intensities.append(standard_data_dictionary.get(constants.AVGAREA))
            concentrations.append(standard_data_dictionary.get(constants.CONC))
            weight = get_weight(standard_data_dictionary, standard_data_index, std_table_metab)
            weights.append(weight)
        standard_data_index = standard_data_index + 1
    yfit, popt, res, sse, r_square, rmse, suggested_standards_xdata = best_suggested_fit(intensities, concentrations,
                                                    intercept, fType,
                                                    int(fitType[
                                                            constants.DEGREE]),
                                                    weights)
    intensities = suggested_standards_xdata
    return yfit, popt, res, sse, r_square, rmse, intensities, fType, y_unit



def create_biosample_data(metab_name, metab_std, intensity, intercept_m,
                          intercept_s, fType_m, fType_s,
                          yfit, popt_m, popt_s, res, biosamples,
                          standard_data_m, standard_data_s,
                          unit_m, unit_s, normalization):
    # TODO: Refactoring code so that less parameters should be passed
    """
    This method serializes the dataframes into a dictionary custom made
    to render the plot, points of fitted curve and unit of concentration
    in case of normalization

    Args:
        metab_name: Metabolite name for which concentrations need to be calculated.
        metab_std:  Metabolite used as std to normalize concentration
        intensity: intensities of standard points of given metabolite
        intercept_m: boolean values corresponding to intercept
        intercept_s: boolean values corresponding to intercept
        fType_m: type of model, can be linear,exp, poly, loglog or power
        fType_s: type of model, can be linear,exp, poly, loglog or power
        yfit: fitted y values of given metabolite
        popt_m: fitted parameters of given metabolite
        popt_s: fitted parameters of std metabolite
        res: residues of given metabolite
        biosamples: dataframe containing experiment samples
        standard_data_m: standard sample df for given metabolite
        standard_data_s: standard sample df for  std metabolite
        unit_m: unit of concentration for given metabolite
        unit_s: unit of concentration for std metabolite
        normalization: boolean values corresponding to normalization
    Returns:
        bsample: json response for biosample points
        fittedpoints: fitted points to plot standard data
        y_unit: unit of concentration
    """
    fittedpoints = []

    for index_std in range(len(yfit)):
        fittedpoints.append({"x": intensity[index_std], "y": yfit[index_std],
                             "residue": res[index_std]})

    metab_df = create_metab_biosample_data(metab_name, intercept_m, fType_m,
                                           popt_m, biosamples,
                                           standard_data_m, unit_m)
    metab_std_df = create_metab_biosample_data(metab_std, intercept_s, fType_s,
                                               popt_s, biosamples,
                                               standard_data_s, unit_s)
    if normalization:
        norm_df = normalize(metab_std_df, metab_df)
        metab_df  = norm_df.rename(
            columns={old_col_names:new_col_names for old_col_names,new_col_names in zip(norm_df.columns, constants.NORM_DF_COLUMNS)})

        bsample = {}
        bsample = get_updated_biosample_points_with_normalization(metab_name, bsample,
                                                    metab_df)
    else:
        bsample = {}
        bsample = get_updated_biosample_points_without_normalization(metab_name, bsample,
                                                       metab_df)
    return bsample, fittedpoints


def get_updated_biosample_points_without_normalization(metab_name, bsample,
                                                  metab_df):
    """
    This function returns updated biosamples information as dictionary
    response without normalized concentration

    Args:
        metab_name: Metabolite name for which biosample concentrations are
                    calculated
        bsample: biosample dictionary to be updated
        metab_df: biosamples dataframe for the metabolite

    Returns:
        bsample: biosamples dictionary with updated information of biosamples
                for metabolite
    """

    for index, row in metab_df.iterrows():
        # this only displays in point name combobox
        sample_name = row[constants.MQ_SAMPLE_NAME]
        area = np.around(float(row[constants.MQ_INTENSITY]),
                         decimals=constants.INTENSITY_PRES)
        conc = np.around(float(row[constants.BIOSAMPLE_CONC]),
                         decimals=constants.CONC_PRES)

        bsample.update({str(row[constants.MQ_SAMPLE_NAME]) + '_' + str(
            row[constants.MQ_FRAGMENT]):
                            {constants.METABOLITE_NAME: metab_name,
                             constants.FRAGMENT_NAME: row[
                                 constants.MQ_FRAGMENT],
                             constants.SAMPLE_NAME: row[
                                 constants.MQ_SAMPLE_NAME],
                             constants.COHORT_NAME: row[
                                 constants.MQ_COHORT_NAME],
                             constants.POINT_NAME: str(sample_name) + "_" +
                             str(
                                 row[constants.MQ_FRAGMENT]),
                             constants.INTENSITY: "%.6f" % (area),
                             constants.CONC: "%.6f" % (conc),
                             constants.UNT: row[constants.UNIT],
                             constants.POINTTYPE: row[constants.POINT_TYPE]}})
    return bsample


def get_updated_biosample_points_with_normalization(metab_name, bsample,
                                                    metab_df):
    """
    This function returns updated biosamples information as dictionary
    response along with normalized concentration

    Args:
        metab_name: Metabolite name for which biosample concentrations are
                    calculated
        bsample: biosample dictionary to be updated
        metab_df: biosamples dataframe for the metabolite

    Returns:
        bsample: biosamples dictionary with updated information of biosamples
                for metabolite
    """

    for index, row in metab_df.iterrows():
        # this only displays in point name combobox
        sample_name = row[constants.MQ_SAMPLE_NAME]
        area = np.around(float(row[constants.MQ_INTENSITY]),
                         decimals=constants.INTENSITY_PRES)
        conc = np.around(float(row[constants.BIOSAMPLE_CONC]),
                         decimals=constants.CONC_PRES)
        norm_conc = np.around(float(row[constants.NORMALIZED_CONC]),
                              decimals=constants.CONC_PRES)

        bsample.update({str(row[constants.MQ_SAMPLE_NAME]) + '_' + str(
            row[constants.MQ_FRAGMENT]):
                            {constants.METABOLITE_NAME: metab_name,
                             constants.FRAGMENT_NAME: row[
                                 constants.MQ_FRAGMENT],
                             constants.SAMPLE_NAME: row[
                                 constants.MQ_SAMPLE_NAME],
                             constants.COHORT_NAME: row[
                                 constants.MQ_COHORT_NAME],
                             constants.POINT_NAME: str(sample_name) + "_" + str(
                                 row[constants.MQ_FRAGMENT]),
                             constants.INTENSITY: "%.6f" % (area),
                             constants.CONC: "%.6f" % (conc),
                             constants.NORMALIZED_CONC: "%.6f" % (norm_conc),
                             constants.UNT: row[constants.UNIT],
                             constants.POINTTYPE: row[constants.POINT_TYPE]}})
    return bsample


def get_curve_fit_response_obj(popt, fType, y_unit, r_square, rmse,
                               fittedPoints, bioSampleJson, sse,
                               new_plot_conc, new_plot_intensity, intercept):
    """
    This function returns curvefit response as dictionary

    Args:
        popt: fitted parameters of given metabolite
        fType: type of model, can be linear,exp, poly, loglog or power
        y_unit: unit of concentration for given metabolite
        r_square: r squared values
        rmse: root mean square error
        fittedPoints: fitted standard points from curvefit
        bioSampleJson: biosample points response as dictionary
        sse: sum of square error
        new_plot_conc: concentration points for plot
        new_plot_intensity: intensity points for plot
        intercept: boolean value corresponding to intercept

    Returns:
            curveFitInfo as a dictionary with information needed for curvefit.
    """
    return {constants.CURVE_FIT_INFO: {
        constants.EQUATION: equation_to_string(popt, fType, intercept),
        constants.Y_UNIT: y_unit,
        constants.GOOD_OF_FIT: {constants.SSE: np.around(sse, 4),
                                constants.R_SQUARE: np.around(r_square, 4),
                                constants.RMSE: np.around(rmse, 4)},
        constants.FIT_POINTS: fittedPoints,
        constants.PLOT_INT: new_plot_intensity,
        constants.PLOT_CONC: new_plot_conc},
            constants.BIOSAMPLES: bioSampleJson}


def normalize(metab_std_df, metab_df):
    """
    This function normalize concentration of metabolite with respect to
    concentration of standard metabolite in that sample.
    :param metab_std_df: standard metabolite dataframe
    :param metab_df: metabolite dataframe
    :return: norm: list of normalised concentration of metabolites in each sample.
    """
    combined_df = metab_df.merge(metab_std_df, on='Sample Name', how='left')
    combined_df[constants.NORM_CONC] = combined_df[constants.METAB_CONC] / combined_df[constants.STD_CONC]
    return combined_df[constants.METAB_DF_COLUMNS]


def create_metab_biosample_data(metab_name, intercept, fType, popt, biosamples,
                                standard_data, unit):
    """
    Return biosample dataframe with calculated concentrations for the input
    metabolite name

    Args:
        metab_name: Metabolite name for which concentrations need to be calculated.
        intercept: boolean values corresponding to intercept
        fType: type of model, can be linear,exp, poly, loglog or power
        popt: fitted parameters of given metabolite
        biosamples: dataframe containing experiment samples
        standard_data: standard sample df for given metabolite
        unit: unit of concentration for given metabolite

    Returns:
        metab_df: biosamples dataframe with corresponding concentration values
    """
    biosamples = intensity_to_conc(metab_name, biosamples, unit, popt, fType,
                                   intercept)
    biosamples = biosamples.loc[:, ~biosamples.columns.duplicated()]

    biosamples = add_point_type(standard_data, biosamples)

    metab_df = biosamples[biosamples[constants.PARENT] == metab_name]
    return metab_df


def add_point_type(standard_data, biosamples):
    """
    Returns a dataframe with a new Point Type column grouping the
    concentrations

    Args:
        standard_data: standard dataframe for the metabolite
        biosamples: biosamples dataframe

    Returns:
        biosamples: biosamples dataframe with point type information
    """
    conc_col = []
    for dct in standard_data:
        if not dct[constants.REJECT]:
            conc_col.append(dct[constants.CONC])
    mincon = min(conc_col)
    maxcon = max(conc_col)
    biosamples[constants.POINT_TYPE] = biosamples[
        constants.BIOSAMPLE_CONC].apply(
        lambda
            x: constants.BQL if x <= mincon else constants.AQL if x > maxcon
            else "In range")
    return biosamples


def get_points_for_plot(popt, fType, maximum_intensity, intercept):
    """
    This function gives points which are used to plot the graph.

    :param popt: equation to get the points
    :param fType: fit type
    :param maximum_intensity: maximum_intensity in the bio sample

    :return:
            plot_intesity : intensity points
            plot_conc : concentration points
    """

    plot_intensity = np.linspace(constants.INITIAL_POINT,
                                 float(maximum_intensity),
                                 num=constants.NUMBER_OF_POINTS)
    minimum_intensity = plot_intensity[1]
    plot_intensity = np.unique(np.append(np.linspace(constants.INITIAL_POINT,
                                        float(minimum_intensity),
                                        constants.NUMBER_OF_POINTS),
                                        plot_intensity))
    plot_conc = interpolation(plot_intensity, popt, fType, intercept)

    round_off_plot_intensity = []
    round_of_plot_conc = []

    for intensity in plot_intensity:
        round_off_plot_intensity.append("%.6f" % intensity)

    for conc in plot_conc:
        round_of_plot_conc.append("%.6f" % conc)
    return round_of_plot_conc, round_off_plot_intensity


def calculate_fit_distance(res, xdata, ydata):
    """
        This function calculates the distance for each standard point from
        the curve and sort them in descending order.
        Args:
            res: It is the distance of all standards from the curve
            x_data : intensity values of the standard points
            y_data : concentration values of the standar points
        Return :
            distances : distance of each point in descending order
            x_data : intensity values of the standard points
            y_data : concentration values of the standar points
    """
    distances = abs(res)
    distances, xdata, ydata = zip(*sorted(zip(distances, xdata, ydata)))
    return xdata, ydata


def delete_standard_data(index, xdata, ydata, weights):
    """
        This function is used to delete a standard.
        Args :
            index  : the index of the standard to be deleted
            x_data : intensity values of the standard points
            y_data : concentration values of the standar points
            weights: weights to apply to the y-coordinates of the sample
            points.
        Return :
            xdata_removed : intensity values with deleting 1 point
            ydata_removed : conc values with deleting 1 point
            weights_removed : weights with 1 weight removed
    """
    xdata = list(xdata)
    ydata = list(ydata)
    weights = list(weights)
    xdata.pop(-1-index)
    ydata.pop(-1-index)
    weights.pop(-1-index)
    return xdata, ydata, weights


def best_suggested_fit(xdata, ydata, intercept, curve_type='linear', degree=1, weights=None):
    """
        This function calculates the best plot for each type of graph
        by removing best suitable standard.
        We calculate the distance for each standard from the curve and
        then remove the standard with maximum distance and recalculate
        R_square value. If its better then remove else nothing.

        return:
            yfit: fitted y values
            popt: fitted parameters
            res: residues
            sse_lin: sum of squares of residuals
            r_square_lin: r squared values
            rmse_lin: root mean squared error
            xdata_del: final xdata with or without deleting standards
    """
    yfit, popt, res, sse, r_square, rmse = curvefit(xdata, ydata, intercept, curve_type, degree, weights)
    xdata, ydata = calculate_fit_distance(res, xdata, ydata)
    xdata_del = xdata
    for trial in range(constants.MAX_STANDARD_CHECK):
        xdata_removed, ydata_removed, weights_removed = delete_standard_data(trial, xdata, ydata, weights)
        yfit_del, popt_del, res_del, sse_del, r_square_del, rmse_del = curvefit(xdata_removed, ydata_removed, intercept, curve_type, degree, weights_removed)
        if r_square_del > r_square:
            yfit, popt, res, sse, r_square, rmse = yfit_del, popt_del, res_del, sse_del, r_square_del, rmse_del
            xdata_del = xdata_removed
    return yfit, popt, res, sse, r_square, rmse, xdata_del
